/*
 * utilHash.c: Functions for managing hash tables. These have been shown
 * experimentally to build hash tables with a dispersion of between .33 (when the
 * table is nearly full (25% free space) and .66. A dispersion of 1 would mean that
 * access is constant time -- with no collisions.
 *
 * These functions call auxilliary functions in utilHashAux.c
 * 
 *
 * Modified Jul 24, 97 to add a function to return the hash key
 */

#include "utilHash.h"
#include <malloc.h>
#include <limits.h>

/* PRIMITIVE: CREATE: returns the table or NULL on failure*/
hashTable createHashTable(initSize, initKeyFunc, shouldFree)
int initSize;
char *(*initKeyFunc)();
int shouldFree;
  {
    return createHashTableSpecialMem(initSize, initKeyFunc, shouldFree, malloc, free);
  }

/* PRIMITIVE: CREATE - SPECIAL MEMORY: creates a table with
 * a special memory manager
 */
hashTable createHashTableSpecialMem(initSize, initKeyFunc, shouldFree, mallocFunc, freeFunc)
int initSize;
char *(*initKeyFunc)();
int shouldFree;
char *(*mallocFunc)();
void (*freeFunc)();
{
    hashTable newTable;
    void (*realFreeFunc)();
    char *(*realMallocFunc)();

    realFreeFunc = (freeFunc == NULL) ? DEFAULT_FREE : freeFunc;
    realMallocFunc = (mallocFunc == NULL) ? DEFAULT_MALLOC : mallocFunc;
    
    newTable = (hashTable)realMallocFunc(sizeof(struct hashHeader));
    if(newTable == NULL) return(NULL);
    newTable->keyFunc = initKeyFunc;
    newTable->mallocFunc = mallocFunc;
    newTable->freeFunc = freeFunc;
    
    if(privateInitHashElements(newTable, initSize) == 0)
      {
        realFreeFunc(newTable);
        return(NULL);
      }
    newTable->numHashed = 0;
    newTable->numUsed = 0;
    newTable->shouldFree = shouldFree;
    return(newTable);
}

/* PRIMITIVE: ADD: returns already present item, added item, or NULL on failure */
itemPtr addToHashTable(theTable, item)
hashTable theTable;
itemPtr item;
  {
    int testMax;
    
    /* see if we need to increase the size */
    testMax = theTable->numUsed;
    testMax+= (testMax >>1) + 1;

    if(theTable->size < testMax)
      {
        int oldSize, oldNumHashed, i;
        itemPtr *oldElements, *currentElement;
        void (*realFreeFunc)();

        realFreeFunc = (theTable->freeFunc == NULL) ? DEFAULT_FREE : theTable->freeFunc;

        /* back up old table information */
        oldSize = theTable->size;
        oldElements = theTable->elements;
        oldNumHashed = theTable->numHashed;

        /* initialize with a bigger table */
        if(! privateInitHashElements(theTable,((theTable->size) + 1) * 2 -1)) return(NULL);
        theTable->numHashed = 0;
        theTable->numUsed = 0;

        /* copy the values from the old to the new table */
        currentElement = oldElements;
        for(i=0; i<oldSize; i++)
          {
            if(*currentElement != NULL && *currentElement != HASH_DELETED)
              {
                oldNumHashed --;
                privateInsert(theTable, *currentElement);
                if(oldNumHashed == 0) break;
              }
            currentElement++;
          }
        realFreeFunc(oldElements);
      }

    return(privateInsert(theTable, item));
  }

/* PRIMITIVE: FIND: returns item if present, otherwise NULL */
itemPtr findInHashTable(theTable, searchString)
hashTable theTable;
char *searchString;
{
  itemPtr *foundAddr, tempElement;
  tempElement = NULL;
  
  foundAddr = privateFindAux(theTable, searchString, 1, &tempElement, 0);

  if(foundAddr == NULL || *foundAddr == HASH_DELETED) return(NULL);

  return(*foundAddr);
}

/* PRIMITIVE: DELETE: returns true if deleted; false if never present or error */
int deleteFromHashTable(theTable, item)
hashTable theTable;
itemPtr item;
{
  itemPtr *addrToDelete, tempElement;
  char *(*realKeyFunc)();
  void (*realFreeFunc)();

  tempElement = NULL;

  realKeyFunc = (theTable->keyFunc == NULL) ? DEFAULT_KEYFUNC : theTable->keyFunc;
  realFreeFunc = (theTable->freeFunc == NULL) ? DEFAULT_FREE : theTable->freeFunc;

  addrToDelete = privateFindAux(theTable, realKeyFunc(item), 1, &tempElement, 0);

  if(addrToDelete == NULL || *addrToDelete == NULL || *addrToDelete == HASH_DELETED)
    {
      return(0);
    }
  if(theTable->shouldFree) realFreeFunc(*addrToDelete);
  *addrToDelete = HASH_DELETED;

  theTable->numHashed--; /* numUsed is deliberately not decremented */
  return(1);
}

/* PRIMITIVE: DESTROY */
void destroyHashTable(theTable)
hashTable theTable;
{
  void (*realFreeFunc)();

  realFreeFunc = (theTable->freeFunc == NULL) ? DEFAULT_FREE : theTable->freeFunc;

  if(theTable->shouldFree)
    hashTableDo(theTable, realFreeFunc, NULL);
  realFreeFunc(theTable->elements);
  theTable->elements = NULL;
  theTable->size = 0;
  theTable->numHashed = 0;
  realFreeFunc(theTable);

}

/* PRIMITIVE: MAP executes a function once for every valid element - there is
  no particular order, and the order can change between calls. Returns the number
  of elements  */
int hashTableDo(theTable, aFunc, extraArg)
hashTable theTable;
void (*aFunc)(); /* takes an item (pointer) as first arg and an arbitrary second arg */
char *extraArg; /* arbitrary second arg to be passed to aFunc */
  {
    int i;
    int howMany = 0;
    itemPtr *currentItem;

    currentItem = theTable->elements;
    for(i=0; i<theTable->size; i++)
      {
        if(*currentItem != NULL && *currentItem !=HASH_DELETED)
          {
            howMany++;
            (aFunc)(*currentItem,extraArg);
          }
        currentItem++;
      }
    return(howMany);
  }

/* PRINTSTRINGS: outputs a list of items in the table - assumes items are strings */
printStrings(theTable)
hashTable theTable;
  {
    printf("{");
    hashTableDo(theTable, printString, NULL);
    printf("}\n");
  }

/* PRIMITIVE DEBUGPRINT: Prints elements, including null and deleted. Assumes */
void debugPrint(theTable)
hashTable theTable;
  {
    char *string;
    int i;

    printf("\n");
    if((theTable->size) < 100) for(i=0; i<theTable->size; i++)
      {
        if((theTable->elements)[i]==NULL) string = "N";
        else if((theTable->elements)[i]==HASH_DELETED) string = "D";
        else string = (char *)testGetString((theTable->elements)[i]);
        
        printf("[%d,%s].",i, string);
      }
    printf("*HASHED: %d USED: %d SIZE: %d DISPERSION: %f\n", theTable->numHashed, theTable->numUsed, theTable->size, dispersion(theTable));
  } 

/* PRIMITIVE SIZE */
int sizeOfHashTable(theTable)
hashTable theTable;
{
  return(theTable->size);
}

/* PRIMITIVE NUMHASHED */
int numHashedInHashTable(theTable)
hashTable theTable;
{
  return(theTable->numHashed);
}

/* PRIMITIVE DISPERSION */
float dispersion(theTable)
hashTable theTable;
  {
    int inRun; /*1 if in a sequence of values or deletes, zero otherwise */
    int changes = 0;
    int i;

    /* we can only have runs if there are more than two elements */
    if(theTable->numUsed <2) return(1.0);

    inRun=((theTable->elements)[(theTable->size)-1]) != NULL;

    for(i=0;i<theTable->size;i++)
      {
        if((theTable->elements)[i] == NULL)
          {
            if(inRun) {changes++; inRun = 0;}
          }
        else
          {
            if(!inRun) {changes++; inRun = 1;}
          }
      }
    return((float)changes / (float)(theTable->numUsed) /2.0);
  }

/* PRIMITIVE: HASH VALUE: returns a hash value for a string - this is
merely a utility - not used internally by this package */
int hashValueOfString(searchString)
char *searchString;
{
  return privateHashVal(searchString, 6, INT_MAX);
}

/* UTILITY for when the items we are entering are just character strings with no other data.
   This is an example of what could be passed to createHashTable as the second argument */
char *testGetString(item)
itemPtr item;
{
  return((char *)item);
}

/* UTILITY to initialize the random number generator from the millisecond timer */
seedRandom()
{
  struct timeval theTime;
  struct timezone theZone;

  gettimeofday(&theTime, &theZone);
  srandom((int)(theTime.tv_usec));
}

/* UTILITY to generate a random string -- useful for testing the hash table routines */
char *randomString()
{
   long size;
   char *theString;
   int i;

   size = (random() % 6) + (random() % 4) + 2;
   theString=(char *)malloc(size+1);
   if(theString==NULL) return(NULL);
   for(i=0;i<size;i++)
     {
       theString[i]='a'+(char)(random()%26);
     }
   theString[size]='\0';

   return(theString);
}

/* UTILITY to print a string followed by a space */
void printString(theString, extraArg)
char *theString;
char *extraArg;
  {
    puts(theString);
    putchar(' ');
  }
